/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.office.odf;

import java.util.HashMap;
import java.util.List;

import org.odftoolkit.odfdom.IElementWriter;
import org.odftoolkit.odfdom.dom.OdfSchemaDocument;
import org.odftoolkit.odfdom.dom.OdfSettingsDom;
import org.odftoolkit.odfdom.dom.OdfSettingsNamespace;
import org.odftoolkit.odfdom.dom.element.office.OfficeDocumentSettingsElement;
import org.odftoolkit.odfdom.pkg.NamespaceName;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

/**
 * The DOM repesentation of the ODF Settings.xml file of an ODF document.
 */
public class Settings extends OdfSettingsDom {

	private static final long serialVersionUID = 766167617530147885L;

	private OfficeSettings officeSettings;

	/**
	 * Creates the DOM representation of an XML file of an Odf document.
	 *
	 * @param odfDocument   the document the XML files belongs to
	 * @param packagePath   the internal package path to the XML file
	 */
	public Settings(OdfSchemaDocument odfDocument, String packagePath) throws SAXException {
		super(odfDocument, packagePath);
	}

	/** Might be used to initialize specific XML Namespace prefixes/URIs for this XML file*/
	@Override
	protected void initialize() throws SAXException {
		for (NamespaceName name : OdfSettingsNamespace.values()) {
			mUriByPrefix.put(name.getPrefix(), name.getUri());
			mPrefixByUri.put(name.getUri(), name.getPrefix());
		}
		final XMLReader xmlReader = mPackage.getXMLReader();
		super.initialize(new SettingsHandler(this, xmlReader), xmlReader, this);
		OfficeDocumentSettingsElement rootElement = this.getRootElement();
		if(rootElement == null){		
			rootElement = new OfficeDocumentSettingsElement(this);
			this.appendChild(rootElement);
			rootElement.appendChild(new OfficeSettings(this));
		}
	}

	/**
	 * Retrieves the Odf Document
	 *
	 * @return The <code>OdfDocument</code>
	 */
	@Override
	public OdfSchemaDocument getDocument() {
		return (OdfSchemaDocument) mPackageDocument;
	}

	/**
	 * @return The root element <office:document-settings> of the settings.xml file as <code>OfficeDocumentSettingsElement</code>.
	 */
	@Override
	public OfficeDocumentSettingsElement getRootElement() {
		return (OfficeDocumentSettingsElement) getDocumentElement();
	}

	public ConfigItemSet getConfigItemSet(String name, boolean forceCreate) {
		if(name==null) {
			return null;
		}
		final OfficeSettings os = getOfficeSettings(forceCreate);
		if(os==null) {
			return null;
		}
		return os.getConfigItem(name, forceCreate);
	}

	public OfficeSettings getOfficeSettings(boolean forceCreate) {
		if(officeSettings==null&&forceCreate) {
			officeSettings = new OfficeSettings(this);
			getRootElement().appendChild(officeSettings);
		}
		return officeSettings;
	}

	public ConfigItemMapEntry getGlobalViewSettings(boolean forceCreate) {
    	final ConfigItemSet configItemSet = getConfigItemSet("ooo:view-settings", forceCreate);
    	if(configItemSet==null) {
    		return null;
    	}
    	final HashMap<String, IElementWriter> configItems = configItemSet.getItems();
        IElementWriter views = configItems.get("Views");
        if(!(views instanceof ConfigItemMapIndexed)) {
        	if(!forceCreate) {
        		return null;
        	}
        	views = new ConfigItemMapIndexed("Views");
        	configItems.put("Views", views);
        }
		final List<ConfigItemMapEntry> viewEntries = ((ConfigItemMapIndexed)views).getEntries();
		if(viewEntries.isEmpty()) {
			if(!forceCreate) {
				return null;
			}
			viewEntries.add(new ConfigItemMapEntry());
		}
		return viewEntries.get(0);
	}

	public ConfigItemMapNamed getViewTables(ConfigItemMapEntry globalViewSettings, boolean forceCreate) {
		IElementWriter tables = globalViewSettings.getItems().get("Tables");
		if(!(tables instanceof ConfigItemMapNamed)) {
			if(!forceCreate) {
				return null;
			}
			tables = new ConfigItemMapNamed("Tables");
			globalViewSettings.getItems().put("Tables", tables);
		}
		return (ConfigItemMapNamed)tables;
	}

	public ConfigItemMapEntry getViewSettings(ConfigItemMapEntry globalViewSettings, String sheetName, boolean forceCreate) {
		if(sheetName==null||sheetName.isEmpty()) {
			return null;
		}
		final ConfigItemMapNamed tables = getViewTables(globalViewSettings, forceCreate);
		if(tables==null) {
			return null;
		}
		ConfigItemMapEntry sheetViewSettings = tables.getItems().get(sheetName);
		if(!(sheetViewSettings instanceof ConfigItemMapEntry)&&forceCreate) {
			sheetViewSettings = new ConfigItemMapEntry(sheetName);
			tables.getItems().put(sheetName, sheetViewSettings);
		}
		return sheetViewSettings;
	}

	public String getActiveSheet() {
		final ConfigItemSet configItemSet = getConfigItemSet("ooo:view-settings", false);
		if(configItemSet!=null) {
			final IElementWriter views = configItemSet.getItems().get("Views");
			if(views instanceof ConfigItemMapIndexed) {
				final List<ConfigItemMapEntry> itemMapEntries = ((ConfigItemMapIndexed)views).getEntries();
				if(!itemMapEntries.isEmpty()) {
					final ConfigItemMapEntry itemMapEntry = itemMapEntries.get(0);
					final IElementWriter activeSheetItem = itemMapEntry.getItems().get("ActiveTable");
					if(activeSheetItem instanceof ConfigItem) {
						return ((ConfigItem)activeSheetItem).getValue();
					}
				}
			}
		}
		return null;
	}

	public void setActiveSheet(String activeSheet) {
		final ConfigItemSet configItemSet = getConfigItemSet("ooo:view-settings", false);
		if(configItemSet!=null) {
			final IElementWriter views = configItemSet.getItems().get("Views");
			if(views instanceof ConfigItemMapIndexed) {
				final List<ConfigItemMapEntry> itemMapEntries = ((ConfigItemMapIndexed)views).getEntries();
				if(!itemMapEntries.isEmpty()) {
					final ConfigItemMapEntry itemMapEntry = itemMapEntries.get(0);
					final IElementWriter activeSheetItem = itemMapEntry.getItems().get("ActiveTable");
					if(activeSheetItem instanceof ConfigItem) {
						((ConfigItem)activeSheetItem).setValue(activeSheet);
					}
				}
			}
		}
	}
}
